##------------------------------------------------------------------------------##
## File:        rstar.stage3.R
##
## Description: This file runs the model in the third stage of the LW estimation.
##------------------------------------------------------------------------------##
rstar.stage3 <- function(log.output,
                         inflation,
                         relative.oil.price.inflation,
                         relative.import.price.inflation,
                         real.interest.rate,
                         covid.dummy,
                         lambda.g,
                         lambda.z,
                         T.og0.omit,
                         sample.end,
                         a.r.constraint=NA,
                         b.y.constraint=NA,
                         run.se=FALSE,
                         xi.00=NA, P.00=NA) {

  stage <- 3

  ## Data must start 8 quarters before estimation period
  T <- length(log.output) - 8

  # Estimate log-linear trend for GDP through 19q4
  ## OLS: Bhat = (X'X)^{-1}X'Y
  x.og <- cbind(rep(1,T+4), 1:(T+4), c(rep(0,56),1:(T+4-56)), c(rep(0,142),1:(T+4-142)))
  y.og <- log.output[5:(T+8)]
  b.og <- solve(t(x.og[1:(nrow(x.og)-T.og0.omit),]) %*% x.og[1:(nrow(x.og)-T.og0.omit),], t(x.og[1:(nrow(x.og)-T.og0.omit),]) %*% y.og[1:(nrow(x.og)-T.og0.omit)])
  # q(t): trend in output estimated through 19q4 and extended through current period
  # q = X*Bhat, where Bhat is estimated through 19q4
  q <- x.og %*% b.og

  # IS curve
  ## Estimate by nonlinear LS the equation:
  ### y(t) - q(t) = phi*d(t) + a1*(y(t-1) - q(t-1) - phi*d(t-1)) + a2(y(t-2) - q(t-2) - phi*d(t-2)) + eps(t)
  y.is    <-  (log.output[9:(T+8)] - q[5:(T+4)]) * 100
  y.is.l1 <-  (log.output[8:(T+7)] - q[4:(T+3)]) * 100
  y.is.l2 <-  (log.output[7:(T+6)] - q[3:(T+2)]) * 100
  d       <-  covid.dummy[9:(T+8)]
  d.l1    <-  covid.dummy[8:(T+7)]
  d.l2    <-  covid.dummy[7:(T+6)]
  ir.is   <-  (real.interest.rate[8:(T+7)] + real.interest.rate[7:(T+6)])/2
  # If the start date is after 2020:Q1, run the model with COVID dummy included
  # Otherwise, initialize phi at zero
  if (T.og0.omit > 0) {
    nls.is   <- nls(y.is ~ phi*d + a_1*(y.is.l1 - phi*d.l1) + a_2*(y.is.l2 - phi*d.l2) +
                    a_r*ir.is + a_0*rep(1,T), start = list(phi=0,a_1=0,a_2=0,a_r=0,a_0=0))
    b.is     <- coef(nls.is)
  } else {
    nls.is   <- nls(y.is ~ a_1*(y.is.l1) + a_2*(y.is.l2) +
                      a_r*ir.is + a_0*rep(1,T), start = list(a_1=0,a_2=0,a_r=0,a_0=0))
    b.is     <- coef(nls.is)
    b.is["phi"] <- 0
  }
  r.is    <-  y.is - predict(nls.is) # residuals
  s.is    <-  sqrt(sum(r.is^2) / (length(r.is)-(length(b.is))))

  # Phillips curve
  ## Estimate by LS, modified to include the covid dummy:
  ### pi(t) = B(L)pi(t-1) + b_y*(y(t-1)-q(t-1)-phi*d(t-1)) + eps_pi(t)
  y.ph <- inflation[9:(T+8)]
  x.ph <- cbind(inflation[8:(T+7)],
                (inflation[7:(T+6)]+inflation[6:(T+5)]+inflation[5:(T+4)])/3,
                (inflation[4:(T+3)]+inflation[3:(T+2)]+inflation[2:(T+1)]+inflation[1:T])/4,
                y.is.l1 - b.is["phi"]*d.l1,
                relative.oil.price.inflation[8:(T+7)],
                relative.import.price.inflation[9:(T+8)])
  b.ph <- solve(t(x.ph) %*% x.ph, t(x.ph) %*% y.ph)
  r.ph <- y.ph - x.ph %*% b.ph
  s.ph <- sqrt(sum(r.ph^2) / (T-dim(x.ph)[2]))

  initial.parameters <- c(b.is["a_1"], b.is["a_2"], b.is["a_r"], b.ph[1:2], b.ph[4:6], 1, s.is, s.ph, 0.7, b.is["phi"])

  ## Initialization of state vector for Kalman filter using HP trend of log output
  ## Pulled into unpack.parameters.stage3.R
  b.pot <- solve(t(x.og) %*% x.og, t(x.og) %*% y.og)
  g.pot <- x.og %*% solve(t(x.og) %*% x.og, t(x.og) %*% y.og)
  xi.00.gpot <- c(100*g.pot[5:3],100*b.pot[2],100*b.pot[2],0,0)

  y.data <- cbind(100 * log.output[9:(T+8)],
                  inflation[9:(T+8)])
  x.data <- cbind(100 * log.output[8:(T+7)],
                  100 * log.output[7:(T+6)],
                  real.interest.rate[8:(T+7)],
                  real.interest.rate[7:(T+6)],
                  inflation[8:(T+7)],
                  (inflation[7:(T+6)]+inflation[6:(T+5)]+inflation[5:(T+4)])/3,
                  (inflation[4:(T+3)]+inflation[3:(T+2)]+inflation[2:(T+1)]+inflation[1:T])/4,
                  relative.oil.price.inflation[8:(T+7)],
                  relative.import.price.inflation[9:(T+8)],
                  covid.dummy[9:(T+8)],
                  covid.dummy[8:(T+7)],
                  covid.dummy[7:(T+6)]) # covid dummy and two lags

  theta.lb <- c(rep(-Inf,length(initial.parameters)))
  theta.ub <- c(rep(Inf,length(initial.parameters)))

  ## Set a lower bound for the Phillips curve slope (b_2) of b.y.constraint, if not NA
  if (!is.na(b.y.constraint)) {
      print(paste("Setting a lower bound of b_y >",as.character(b.y.constraint),"in Stage 3"))
      if (initial.parameters[6] < b.y.constraint) {
          initial.parameters[6] <- b.y.constraint
      }
      theta.lb[6] <- b.y.constraint
  }

  ## Set an upper bound for the IS curve slope (a_3) of a.r.constraint, if not NA
  if (!is.na(a.r.constraint)) {
      print(paste("Setting an upper bound of a_r <",as.character(a.r.constraint),"in Stage 3"))
      if (initial.parameters[3] > a.r.constraint) {
          initial.parameters[3] <- a.r.constraint
      }
      theta.ub[3] <- a.r.constraint
  }

  P.00.gpot <- calculate.covariance(initial.parameters, theta.lb, theta.ub, y.data, x.data, stage, lambda.g, lambda.z, xi.00=xi.00.gpot)

  f <- function(theta) {return(-log.likelihood.wrapper(theta, y.data, x.data, stage,
                                                       lambda.g, lambda.z,
                                                       xi.00=NA, P.00=NA,
                                                       xi.00.gpot, P.00.gpot)$ll.cum)}
  nloptr.out <- nloptr(initial.parameters, f, eval_grad_f=function(x) {gradient(f, x)},
                       lb=theta.lb,ub=theta.ub,
                       opts=list("algorithm"="NLOPT_LD_LBFGS","xtol_rel"=1.0e-8,"maxeval"=200))
  theta <- nloptr.out$solution

  if (nloptr.out$status==-1 | nloptr.out$status==5) {
      print("Look at the termination conditions for nloptr in Stage 3")
      stop(nloptr.out$message)
  }

  ## Get xi.00 and P.00
  init.vals <- unpack.parameters.stage3(theta, y.data, x.data, lambda.g, lambda.z,
                                        xi.00=NA, P.00=NA, xi.00.gpot, P.00.gpot)
  xi.00 <- init.vals$xi.00
  P.00  <- init.vals$P.00
  print(paste("GLS switch value:",init.vals$gls))

  log.likelihood <- log.likelihood.wrapper(theta, y.data, x.data, stage, lambda.g, lambda.z, xi.00, P.00)$ll.cum
  states <- kalman.states.wrapper(theta, y.data, x.data, stage, lambda.g, lambda.z, xi.00, P.00)

  ## If run.se = TRUE, compute standard errors for estimates of the states and report run time
  if (run.se) {
      ptm <- proc.time()
      se <- kalman.standard.errors(T, states, theta, y.data, x.data, stage, lambda.g, lambda.z, xi.00, P.00, niter, a.r.constraint, b.y.constraint, sample.end)
      print("Standard error procedure run time")
      print(proc.time() - ptm)
  }

  ## One-sided (filtered) estimates
  trend.filtered      <- states$filtered$xi.tt[,4] * 4
  z.filtered          <- states$filtered$xi.tt[,6]
  rstar.filtered      <- trend.filtered * theta[9] + z.filtered
  potential.filtered  <- states$filtered$xi.tt[,1]/100
  output.gap.filtered <- y.data[,1] - (potential.filtered * 100) - theta[length(theta)]*covid.dummy[9:(T+8)] # with dummy term

  ## Two-sided (smoothed) estimates
  trend.smoothed      <- states$smoothed$xi.tT[,4] * 4
  z.smoothed          <- states$smoothed$xi.tT[,6]
  rstar.smoothed      <- trend.smoothed * theta[9] + z.smoothed
  potential.smoothed  <- states$smoothed$xi.tT[,1]/100
  output.gap.smoothed <- y.data[,1] - (potential.smoothed * 100) - theta[length(theta)]*covid.dummy[9:(T+8)] # with dummy term

  ## Save variables to return
  return.list <- list()
  return.list$rstar.filtered      <- rstar.filtered
  return.list$trend.filtered      <- trend.filtered
  return.list$z.filtered          <- z.filtered
  return.list$potential.filtered  <- potential.filtered
  return.list$output.gap.filtered <- output.gap.filtered
  return.list$rstar.smoothed      <- rstar.smoothed
  return.list$trend.smoothed      <- trend.smoothed
  return.list$z.smoothed          <- z.smoothed
  return.list$potential.smoothed  <- potential.smoothed
  return.list$output.gap.smoothed <- output.gap.smoothed
  return.list$theta               <- theta
  return.list$log.likelihood      <- log.likelihood
  return.list$states              <- states
  return.list$xi.00               <- xi.00
  return.list$P.00                <- P.00
  return.list$y.data              <- y.data
  return.list$initial.parameters  <- initial.parameters
  return.list$init.vals           <- init.vals
  if (run.se) { return.list$se    <- se }
  return(return.list)
}
